@@ -753,6 +753,8 @@ xfical_period get_period(icalcomponent *c_event)
 {
     icalproperty *p;
     xfical_period per;
+    struct icaldurationtype duration_tmp;
+    gint dur_int;
 
     /* Exactly one start time must be there */
     p = icalcomponent_get_first_property(c_event, ICAL_DTSTART_PROPERTY);
@@ -772,6 +774,16 @@ xfical_period get_period(icalcomponent *c_event)
         per.etime = icalproperty_get_dtend(p);
         per.etime = convert_to_local_timezone(per.etime, p);
         per.duration = icaltime_subtract(per.etime, per.stime);
+        if (icaltime_is_date(per.stime)
+            && icaldurationtype_as_int(per.duration) != 0) {
+        /* need to subtract 1 day.
+         * read explanation from appt_add_internal: appt->endtime processing */
+            duration_tmp = icaldurationtype_from_int(60*60*24);
+            dur_int = icaldurationtype_as_int(per.duration);
+            dur_int -= icaldurationtype_as_int(duration_tmp);
+            per.duration = icaldurationtype_from_int(dur_int);
+            per.etime = icaltime_add(per.stime, per.duration);
+        }
     }
     else {
         p = icalcomponent_get_first_property(c_event, ICAL_DURATION_PROPERTY);
@@ -1254,6 +1266,8 @@ char *appt_add_internal(appt_data *appt, gboolean add, char *uid
     struct icaldurationtype duration;
     int i, cnt;
     gchar *byday_a[] = {"MO", "TU", "WE", "TH", "FR", "SA", "SU"};
+    icaltimezone *l_icaltimezone = NULL;
+    const char *text;
 
     dtstamp = icaltime_current_time_with_zone(utc_icaltimezone);
     if (add) {
@@ -1277,6 +1291,7 @@ char *appt_add_internal(appt_data *appt, gboolean add, char *uid
            , icalproperty_new_class(ICAL_CLASS_PUBLIC)
            , icalproperty_new_dtstamp(dtstamp)
            , icalproperty_new_created(create_time)
+           , icalproperty_new_lastmodified(dtstamp)
            , NULL);
 
     if XFICAL_STR_EXISTS(appt->title)
@@ -1301,7 +1316,11 @@ char *appt_add_internal(appt_data *appt, gboolean add, char *uid
 
     if XFICAL_STR_EXISTS(appt->starttime) {
         wtime=icaltime_from_string(appt->starttime);
-        if XFICAL_STR_EXISTS(appt->start_tz_loc) {
+        if (appt->allDay) { /* date */
+            icalcomponent_add_property(ievent
+                    , icalproperty_new_dtstart(wtime));
+        }
+        else if XFICAL_STR_EXISTS(appt->start_tz_loc) {
             if (strcmp(appt->start_tz_loc, "UTC") == 0) {
                 wtime=icaltime_convert_to_zone(wtime, utc_icaltimezone);
                 icalcomponent_add_property(ievent
@@ -1330,7 +1349,15 @@ char *appt_add_internal(appt_data *appt, gboolean add, char *uid
     }
     else if XFICAL_STR_EXISTS(appt->endtime) {
         wtime=icaltime_from_string(appt->endtime);
-        if XFICAL_STR_EXISTS(appt->end_tz_loc) {
+        if (appt->allDay) {
+            /* need to add 1 day. For example:
+             * DTSTART=20070221 & DTEND=20070223
+             * means only 21 and 22 February */
+            duration = icaldurationtype_from_int(60*60*24);
+            wtime = icaltime_add(wtime, duration);
+            icalcomponent_add_property(ievent, icalproperty_new_dtend(wtime));
+        }
+        else if XFICAL_STR_EXISTS(appt->end_tz_loc) {
             if (strcmp(appt->end_tz_loc, "UTC") == 0) {
                 wtime=icaltime_convert_to_zone(wtime, utc_icaltimezone);
                 icalcomponent_add_property(ievent
@@ -1378,7 +1405,26 @@ char *appt_add_internal(appt_data *appt, gboolean add, char *uid
             recur_p += g_sprintf(recur_p, ";COUNT=%d", appt->recur_count);
         }
         else if (appt->recur_limit == 2) { /* needs to be in UTC */
-            recur_p += g_sprintf(recur_p, ";UNTIL=%sZ", appt->recur_until);
+/* BUG 2937: convert recur_until to utc from start time timezone */
+            wtime = icaltime_from_string(appt->recur_until);
+            if XFICAL_STR_EXISTS(appt->start_tz_loc) {
+            /* Null == floating => no special action needed */
+                if (strcmp(appt->start_tz_loc, "floating") == 0) {
+                    wtime = icaltime_convert_to_zone(wtime
+                            , local_icaltimezone);
+                }
+                else if (strcmp(appt->start_tz_loc, "UTC") != 0) {
+                /* FIXME: add this vtimezone to vcalendar if it is not there */
+                    l_icaltimezone = icaltimezone_get_builtin_timezone(
+                            appt->start_tz_loc);
+                    wtime = icaltime_convert_to_zone(wtime, l_icaltimezone);
+                }
+            }
+            else
+                wtime = icaltime_convert_to_zone(wtime, local_icaltimezone);
+            wtime = icaltime_convert_to_zone(wtime, utc_icaltimezone);
+            text  = icaltime_as_ical_string(wtime);
+            recur_p += g_sprintf(recur_p, ";UNTIL=%s", text);
         }
         recur_p2 = recur_p; /* store current pointer */
         for (i = 0, cnt = 0; i <= 6; i++) {
@@ -1501,11 +1547,12 @@ appt_data *xfical_appt_get_internal(char *ical_uid)
     icalproperty *p = NULL;
     gboolean key_found = FALSE;
     const char *text;
-    struct icaltimetype itime, stime, etime, sltime, eltime;
+    struct icaltimetype itime, stime, etime, sltime, eltime, wtime;
+    icaltimezone *l_icaltimezone = NULL;
     icalparameter *itime_tz;
     icalproperty_transp xf_transp;
     struct icalrecurrencetype rrule;
-    struct icaldurationtype duration;
+    struct icaldurationtype duration, duration_tmp;
     int i, cnt, day;
 
     for (c = icalcomponent_get_first_component(ical, ICAL_VEVENT_COMPONENT); 
@@ -1700,6 +1747,7 @@ appt_data *xfical_appt_get_internal(char *ical_uid)
                     case ICAL_CLASS_PROPERTY:
                     case ICAL_DTSTAMP_PROPERTY:
                     case ICAL_CREATED_PROPERTY:
+                    case ICAL_LASTMODIFIED_PROPERTY:
                         break;
                     default:
                         g_warning("xfical_appt_get_internal: unknown property %s"
@@ -1711,19 +1759,45 @@ appt_data *xfical_appt_get_internal(char *ical_uid)
         }
     } 
 
+    if (key_found) {
     /* need to set missing endtime or duration */
-    if (appt.use_duration) { 
-        etime = icaltime_add(stime, duration);
-        text  = icaltime_as_ical_string(etime);
-        g_strlcpy(appt.endtime, text, 17);
-        appt.end_tz_loc = appt.start_tz_loc;
-    }
-    else {
-        duration = icaltime_subtract(eltime, sltime);
-        appt.duration = icaldurationtype_as_int(duration);
-    }
-    if (key_found)
+        if (appt.use_duration) { 
+            etime = icaltime_add(stime, duration);
+            text  = icaltime_as_ical_string(etime);
+            g_strlcpy(appt.endtime, text, 17);
+            appt.end_tz_loc = appt.start_tz_loc;
+        }
+        else {
+            duration = icaltime_subtract(eltime, sltime);
+            appt.duration = icaldurationtype_as_int(duration);
+            if (appt.allDay && appt.duration) {
+        /* need to subtract 1 day.
+         * read explanation from appt_add_internal: appt->endtime processing */
+                duration_tmp = icaldurationtype_from_int(60*60*24);
+                appt.duration -= icaldurationtype_as_int(duration_tmp);
+                duration = icaldurationtype_from_int(appt.duration);
+                etime = icaltime_add(stime, duration);
+                text  = icaltime_as_ical_string(etime);
+                g_strlcpy(appt.endtime, text, 17);
+            }
+        }
+        if (appt.recur_limit == 2) { /* BUG 2937: convert back from UTC */
+            wtime = icaltime_from_string(appt.recur_until);
+            if (! XFICAL_STR_EXISTS(appt.start_tz_loc))
+                wtime = icaltime_convert_to_zone(wtime, local_icaltimezone);
+            else if (strcmp(appt.start_tz_loc, "floating") == 0)
+                wtime = icaltime_convert_to_zone(wtime, local_icaltimezone);
+            else if (strcmp(appt.start_tz_loc, "UTC") != 0) {
+                /* done if in UTC, but real conversion needed if not */
+                l_icaltimezone = icaltimezone_get_builtin_timezone(
+                        appt.start_tz_loc);
+                wtime = icaltime_convert_to_zone(wtime, l_icaltimezone);
+            }
+            text  = icaltime_as_ical_string(wtime);
+            g_strlcpy(appt.recur_until, text, 17);
+        }
         return(&appt);
+    }
     else
         return(NULL);
 }
